var Promise = require('bluebird');
var crypto = require('crypto');
var path = require('path');
var util = require('util');
var os = require('os');
var blockDecoder = require('./BlockDecoderProxy.js');
var Constants = require('./Constants.js');
var elliptic = require('elliptic');
var EC = elliptic.ec;

var grpc = require('grpc');
//------------------------------------
var hfc = require('./Client.js');
var Identity = require('./msp/identity.js').Identity;
var SigningIdentity = require('./msp/identity.js').SigningIdentity;
var Signer = require('./msp/identity.js').Signer;

var CryptoSuite_ECDSA_AES = require('./impl/CryptoSuite_ECDSA_AES.js');
var TransactionID = require('./TransactionID.js');
var ECDSAKey = require('./impl/ecdsa/key.js');
var clientUtils = require('./client-utils.js');
var utils = require('./utils.js');
var Peer = require('./Peer.js');
var Orderer = require('./Orderer.js');
var ChannelEventHub = require('./ChannelEventHub.js');
var jsrsa = require('jsrsasign');
var KEYUTIL = jsrsa.KEYUTIL;

var _queryProto = grpc.load(__dirname + '/protos/peer/query.proto').protos;
var _ledgerProto = grpc.load(__dirname + '/protos/common/ledger.proto').common;
var _commonProto = grpc.load(__dirname + '/protos/common/common.proto').common;
var _transProto = grpc.load(__dirname + '/protos/peer/transaction.proto').protos;
var _proposalProto = grpc.load(__dirname + '/protos/peer/proposal.proto').protos;
var _configtxProto = grpc.load(__dirname + '/protos/common/configtx.proto').common;
var _chaincodeProto = grpc.load(__dirname + '/protos/peer/chaincode.proto').protos;
//------------------------------------


var default_timeout = 20000;


Buffer.prototype.toBytes = function () {
  var bytes = [];
  for (var i = 0; i < this.length; i++) {
    bytes.push(this.readUInt8(i))
  }
  return bytes
};


function tryParseJSON(str) {
  var json = str;
  try {
    json = JSON.parse(str)
  }
  catch (err) {

  }
  return json
};

function formatChaincodeResult(payloadStr) {
  var payload = tryParseJSON(payloadStr);
  var result = {};
  if (typeof payload == 'object' && payload.constructor == Array) {
    result.result = payload
  } else if (typeof payload == 'object') {
    for (var item in payload) {
      result[item] = payload[item]
    }
  } else {
    result.result = payload
  }
  return result
};


function ClientProxy() {
  var client = new hfc();
  this._client = client;
  this._cryptoSuite = new CryptoSuite_ECDSA_AES(256, 'SHA2')
};

/**
 * [setAdminSigningIdentity description]
 * @param {[type]} private_key [description]
 * @param {[type]} certificate [description]
 * @param {[type]} mspid       [description]
 */
ClientProxy.prototype.setAdminSigningIdentity = function (private_key, certificate, mspid) {
  this._client.setAdminSigningIdentity(private_key, certificate, mspid)
};

/**
 * [loadCertificate description]
 * @param  {[type]} certificate [description]
 * @param  {[type]} mspid       [description]
 * @return {[type]}             [description]
 */
ClientProxy.prototype.setUserIdentity = function (certificate, mspid) {
  var self = this;
  self._certificate = certificate;
  self._mspId = mspid;
  self._identity = this._newIdentity(certificate, self._mspId)
};

ClientProxy.prototype.setClientSigningIdentity = function (private_key, certificate, mspid) {
  this._client.setAdminSigningIdentity(private_key, certificate, mspid);
  this.setUserIdentity(certificate, mspid);
  this._private_key = private_key
};

/**
 * [invokeChaincode description]
 * @param  {object} target object   e.g. {peer_url,orderer_url,event_url,peer_opts,orderer_opts}
 * @param  {string} channel_id   [description]
 * @param  {string} chaincode_id [description]
 * @param  {string} fcn          [description]
 * @param  {array} args array of string        [description]
 * @return {promise}  promise of object {proposalResponses:'',chaincode_result:'',tx_result:''}
 */
ClientProxy.prototype.invokeChaincode = function (target_obj, channel_id, chaincode_id, fcn, args, timeout) {
  var self = this;
  if (!target_obj.peer_url) {
    throw new Error('Missing "peer_url" in target_obj');
  }
  var _timeout = timeout || default_timeout;
  var proposal = self.buildTransactionProposal(channel_id, chaincode_id, fcn, args);
  var proposal_bytes = proposal.toBuffer();

  var proposal_signature = self.sign(self._private_key, proposal_bytes);

  var verifyResult = self.verifyProposal(proposal_signature, proposal_bytes);
  if (!verifyResult) {
    return Promise.reject(new Error('invalid signature:call setClientSigningIdentity() again using right privatekey'))
  }

  var context = {};
  return self.sendPeersProposal(target_obj.peer_url, proposal_bytes, proposal_signature, _timeout, target_obj.peer_opts)
    .then(function (responses) {
      self.decodeProposalResponse(responses)
      if( responses[0].chaincode_error ){
        return Promise.reject(responses[0].chaincode_error)
      }
      context.chaincode_result = responses[0].chaincode_result;
      context.proposalResponses = responses;
      if (self.hasWriteSet(responses) && target_obj.orderer_url) {
        var tx_payload_bytes = self.buildTransactionPayload(proposal_bytes, responses);
        return Promise.resolve(tx_payload_bytes)
      } else {
        return Promise.resolve(null)
      }

    })
    .then(function (tx_payload_bytes) {
      if (tx_payload_bytes) {
        var tx_signature = self.sign(self._private_key, tx_payload_bytes);
        return self.sendTransaction(target_obj, tx_payload_bytes, tx_signature, _timeout)
          .then(function (tx_result) {
            context.tx_result = tx_result;
            return Promise.resolve(context)
          })
          .catch(function (err) {
            return Promise.reject(err)
          })

      } else {
        return Promise.resolve(context)
      }
    })

};
/**
 * [createChannel description]
 * @param  {string} order_url
 * @param  {string} channelName
 * @param  {buffer} envelope   [channel文件byte[]]
 * @param  {object} opts  orderer target options  e.g. {pem:'--- tls cert ---'}
 * @param  {Array.ConfigSignature} signatures  signatures of members
 * @return {[type]} [description]
 */
ClientProxy.prototype.createChannelByFile = function (order_url, channelName, envelope, opts,signatures) {
  let channelConfig;
  try {
    channelConfig = this._client.extractChannelConfig(envelope);
    //console.log("channelConfig:",JSON.stringify(_configtxProto.ConfigUpdate.decode(channelConfig)));
  } catch (err) {
    return Promise.reject(err);
  }
  return this.createOrUpdateChannel(order_url, channelName, channelConfig, opts,signatures);
};
/**
 * [createChannel description]
 * @param  {string} order_url
 * @param  {string} channelName
 * @param  {buffer} updateConfig   [channelUpdateConfig对象grpc encode Byte[]]
 * @param  {object} opts  orderer target options  e.g. {pem:'--- tls cert ---'}
 * @param  {Array.ConfigSignature} signatures  signatures of members
 */
ClientProxy.prototype.createOrUpdateChannel = function (order_url, channelName, updateConfig, opts,signatures) {
  let order = new Orderer(order_url, opts);
  let request = {
    name: channelName,
    orderer: order,
    config: updateConfig,
    signatures: signatures,
    txId: this._client.newTransactionID(true)
  };
  return this._client.createChannel(request);
};

/**
 * [joinChannel description]
 * @param  {[type]} target_obj    {peer_url,orderer_url,event_url,peer_opts,orderer_opts}
 * @param  {[type]} channel_id   [description]
 * @param  {[type]} timeout      [description]
 * @return {[type]}              [description]
 */
ClientProxy.prototype.joinChannel = function (target_obj, channel_id, timeout) {
  let _timeout = timeout || default_timeout;
  let channel = this.getChannel(channel_id);
  let orderRequest = {
    orderer: new Orderer(target_obj.orderer_url, target_obj.orderer_opts),
    txId: this._client.newTransactionID(true)
  };
  let targets = [new Peer(target_obj.peer_url, target_obj.peer_opts)];
  // if (!Array.isArray(target_obj.peer_url,)) {
  //   target_obj.peer_url = [target_obj.peer_url];
  // }
  // for (let peer_url of target_obj.peer_url) {
  //   targets.push(new Peer(peer_url, null));
  // }
  return channel.getGenesisBlock(orderRequest).then(genesisBlock => {
    let channelRequest = {
      targets: targets,
      block: genesisBlock,
      txId: this._client.newTransactionID(true)
    };
    return channel.joinChannel(channelRequest, _timeout);
  });
};
/**
 * [getChannelInfo description]
 * @param  {string} peer_url
 * @param  {string} channel_id   [channel名]
 * @param  {object} opts  peer target options  e.g. {pem:'--- tls cert ---'}
 * @return {[type]}              [description]
 */
ClientProxy.prototype.getChannelInfo = function (peer_url, channel_id, opts) {
  var channel = this.getChannel(channel_id);
  var target = new Peer(peer_url, opts);
  return channel.queryInfo(target, true).then(channelInfo => {
    channelInfo.currentBlockHash = channelInfo.currentBlockHash.toBuffer().toString('hex');
    channelInfo.previousBlockHash = channelInfo.previousBlockHash.toBuffer().toString('hex');
    return Promise.resolve(channelInfo);
  })
};

/**
 * [queryBlockByHash description]
 * @param  {string} peer_url
 * @param  {string} channel_id   [channel名]
 * @param  {string} blockHash   [blockHash]
 * @param  {object} opts  peer target options  e.g. {pem:'--- tls cert ---'}
 * @return {[type]}              [description]
 */
ClientProxy.prototype.queryBlockByHash = function (peer_url, channel_id, blockHash, opts) {
  var channel = this.getChannel(channel_id);
  var target = new Peer(peer_url, opts);
  return channel.queryBlockByHash(Buffer.from(blockHash, 'hex'), target, true);
};

/**
 * [queryBlockByNumber description]
 * @param  {string} peer_url
 * @param  {string} channel_id   [channel名]
 * @param  {int} number   [blockNumber]
 * @param  {object} opts  peer target options  e.g. {pem:'--- tls cert ---'}
 * @return {[type]}              [description]
 */
ClientProxy.prototype.queryBlockByNumber = function (peer_url, channel_id, number, opts) {
  var channel = this.getChannel(channel_id);
  var target = new Peer(peer_url, opts);
  return channel.queryBlock(number, target, true);
};

/**
 * [queryTransaction description]
 * @param  {string} peer_url
 * @param  {string} channel_id   [channel名]
 * @param  {string} tx_id   [tx_id]
 * @param  {object} opts  peer target options  e.g. {pem:'--- tls cert ---'}
 * @return {[type]}              [description]
 */
ClientProxy.prototype.queryTransaction = function (peer_url, channel_id, tx_id, opts) {
  var channel = this.getChannel(channel_id);
  var target = new Peer(peer_url, opts);
  return channel.queryTransaction(tx_id, target, true);
};

/**
 * [installChaincode description]
 * @param  {object[]} peerTargets  [{url:'',opts:{pem:'--- tls cert ---'}},]
 * @param  {string} chaincodePath  string  Required.
 * @param  {string} chaincodeId  string  Required. Name of the chaincode
 * @param  {string} chaincodeVersion   string  Required. Version string of the chaincode, such as 'v1'
 * @param  {string[]} channelNames
 * @param  {string} chaincodeType  string  Optional. Type of chaincode. One of 'golang', 'car', 'node' or 'java'. Default is 'golang'. Note that 'java' is not yet supported.
 *
 *return example [[{"version":0,"timestamp":null,"response":{"status":200,"message":"","payload":{"type":"Buffer","data":[79,75]}},"payload":{"type":"Buffer","data":[]},"endorsement":null}]]
 */
ClientProxy.prototype.installChaincode = function (peerTargets, chaincodePath, chaincodeId, chaincodeVersion, channelNames, chaincodeType) {
  console.info(">>>>>>user context:", this._client._userContext);
  let targets = [];
  if (!Array.isArray(peerTargets)) {
    peerTargets = [peerTargets];
  }
  for (let peerTarget of peerTargets) {
    targets.push(new Peer(peerTarget.url, peerTarget.opts));
  }
  let request = {
    targets: targets,
    chaincodePath: chaincodePath,
    chaincodeId: chaincodeId,
    chaincodeVersion: chaincodeVersion,
    chaincodeType: chaincodeType || 'golang',
    channelNames: channelNames
  };
  return this._client.installChaincode(request);
};

/**
 * [instantiateUpgradeChaincode description]
 * @param  {object} target_obj    {peer_url,orderer_url,event_url,peer_opts,orderer_opts}
 * @param  {object} request
 * @param  {string} request.channel_id        Required.
 * @param  {string} request.chaincodeId        Required.
 * @param  {string} request.chaincodeVersion    Required.
 * @param  {string} request.chaincodeType      Optional.
 * @param  {string} request.func               Optional. Default : 'init'.
 * @param  { Array.<String>} request.args      Optional.
 * @param  {boolean} request.isUpdate          Optional. Default: false
 * @param  {Object} request.endorsementPolicy  Optional.
 * @param  {int}   timeout                    Optional.
 * return example 同install
 */
ClientProxy.prototype.instantiateUpgradeChaincode = function (target_obj, request, timeout) {
  let _timeout = timeout || default_timeout;
  let targets = [];
  let channel = this.getChannel(request.channel_id);
  let command = "deploy";
  let tx_id = this.newTransactionID(true);
  targets.push(new Peer(target_obj.peer_url, target_obj.peer_opts));
  let req = {
    targets: targets,
    chaincodeId: request.chaincodeId,
    chaincodeVersion: request.chaincodeVersion,
    chaincodeType: request.chaincodeType || 'golang',
    txId: tx_id,
    fcn: request.func,
    args: request.args,
    "endorsement-policy": request.endorsementPolicy
  };
  if (request.isUpdate === true) {
    command = "upgrade";
  }
  return channel._sendChaincodeProposal(req, command,_timeout).then(results => {
    let proposalResponses = results[0];
    let proposal = results[1];
    //There only one peer.
    if (proposalResponses && proposalResponses[0].response &&
      proposalResponses[0].response.status === 200) {
      logger.info('chain code proposal was success done');
      let tx_payload_bytes = this.buildTransactionPayload(proposal, proposalResponses);
      let tx_signature = this.sign(this._private_key, tx_payload_bytes);
      return this.sendTransaction(target_obj, tx_payload_bytes, tx_signature,_timeout);
    } else {
      logger.error('chain code proposal was failed !');
      return Promise.reject(new Error("sendChainCodeProposal error:" + proposalResponses[0].response.message));
    }
  });
};


/**
 * [queryInstantiatedChaincodes description]
 * @param  {string} peer_url
 * @param  {string} channel_id   [channel名]
 * @param  {object} opts  peer target options  e.g. {pem:'--- tls cert ---'}
 * @return {[type]}              [description]
 */
ClientProxy.prototype.queryInstantiatedChaincodes = function (peer_url, channel_id, opts) {
  var channel = this.getChannel(channel_id);
  var target = new Peer(peer_url, opts);
  return channel.queryInstantiatedChaincodes(target, true);
};

/**
 * [queryInstalledChaincodes description]
 * @param  {string} peer_url
 * @param  {object} opts  peer target options  e.g. {pem:'--- tls cert ---'}
 * @return {[type]}              [description]
 */
ClientProxy.prototype.queryInstalledChaincodes = function (peer_url, opts) {
  var target = new Peer(peer_url, opts);
  return this._client.queryInstalledChaincodes(target, true);
};

/**
 * [queryChaincodeByName description]
 * @param  {string} peer_url
 * @param  {string} channel_id   [channel名]
 * @param  {string} chaincode_id [chaincode名]
 * @param  {object} opts  peer target options  e.g. {pem:'--- tls cert ---'}
 * @return {byte[]} Promise of chaincode (bytes)
 */
ClientProxy.prototype.queryChaincodeByName = function (peer_url, channel_id, chaincode_id, opts) {
  var channel = this.getChannel(channel_id);
  var target = new Peer(peer_url, opts);
  var request = {
    targets: [target],
    chaincodeId: Constants.LSCC,
    fcn: "GetChaincodeData",
    args: [channel_id, chaincode_id]
  };
  return channel.queryByChaincode(request, true).then(chaincodeDatas => {
    if (chaincodeDatas && Array.isArray(chaincodeDatas) && chaincodeDatas.length === 1) {
      return Promise.resolve(chaincodeDatas[0]);
    }
    return Promise.resolve("");
  });
};


/**
 * [init description]
 * @return {[type]} [description]
 */
ClientProxy.setLogger = function (logger) {
  hfc.setLogger(logger)
};


/**
 * 得到 公/私钥 对象
 * @param  {[type]} pem [description]
 * @return {[type]}     [description]
 */
ClientProxy.getKey = function (pem) {
  return KEYUTIL.getKey(pem)
};

/**
 * [_newIdentity description]
 * @param  {[type]} certificate [description]
 * @param  {[type]} mspId       [description]
 * @return {[type]}             [description]
 */
ClientProxy.prototype._newIdentity = function (certificate, mspId) {
  var key = null
  try {
    key = KEYUTIL.getKey(certificate);
  } catch (err) {
    throw new Error('Failed to parse key from PEM: ' + err);
  }

  var theKey = null
  if (key && key.type && key.type === 'EC') {
    theKey = new ECDSAKey(key);
  } else {
    throw new Error('Failed to parse key from PEM: ' + key)
  }

  return new Identity(certificate, theKey, mspId, new CryptoSuite_ECDSA_AES(256, 'SHA2'))
}


/**
 * [newTransactionID description]
 * @param  {boolean} isAdmin [description]
 * @return {[type]} [description]
 */
ClientProxy.prototype.newTransactionID = function (isAdmin) {
  if (!this._identity) {
    throw new Error('load certificate first before call newTransactionID')
  }
  return new TransactionID(this._identity, isAdmin)
}

/**
 * [getClientCertHash description]
 * @return {[type]} [description]
 */
ClientProxy.prototype.getClientCertHash = function () {
  let hash = null;
  if (this._certificate) {
    let der_cert = utils.pemToDER(this._certificate);
    var sha256 = crypto.createHash('sha256');
    hash = sha256.update(der_cert).digest();
  }
  return hash;
}


/**
 * [buildTransactionProposal description]
 * @param  {[type]} channel_id   [description]
 * @param  {[type]} chaincode_id [description]
 * @param  {[type]} fcn           [description]
 * @param  {[type]} args          [description]
 * @return {[type]}               [description]
 */
ClientProxy.prototype.buildTransactionProposal = function (channel_id, chaincode_id, fcn, args) {
  var request = this._buildRequest(channel_id, chaincode_id, fcn, args)
  return this._buildTransactionProposal(request)
}

/**
 * [_buildRequest description]
 * @param  {[type]} channel_id   [description]
 * @param  {[type]} chaincode_id [description]
 * @param  {[type]} fcn           [description]
 * @param  {[type]} args          [description]
 * @return {[type]}               [description]
 */
ClientProxy.prototype._buildRequest = function (channel_id, chaincode_id, fcn, args) {
  var self = this

  var channel = self.getChannel(channel_id)


  if (!channel) {
    let message = util.format('Channel %s was not defined in the connection profile', channelName);
    logger.error(message);
    throw new Error(message);
  }

  var identity = self._identity
  var tx_id = self.newTransactionID();

  // send proposal to endorser
  var request = {
    chaincodeId: chaincode_id,
    fcn: fcn,
    args: args,
    chainId: channel_id,
    txId: tx_id,
    clientCertHash: self.getClientCertHash(self._certificate)
  };
  return request
}

/**
 * [_buildTransactionProposal description]
 * @param  {[type]} request [description]
 * @return {[type]}         [description]
 */
ClientProxy.prototype._buildTransactionProposal = function (request) {
  var identity = this._identity
  var errorMsg = clientUtils.checkProposalRequest(request);

  if (errorMsg) {
    // do nothing so we skip the rest of the checks
  } else if (!request.args) {
    // args is not optional because we need for transaction to execute
    errorMsg = 'Missing "args" in Transaction proposal request';
  }

  if (errorMsg) {
    logger.error('buildTransactionProposal error ' + errorMsg);
    throw new Error(errorMsg);
  }

  var args = [];
  args.push(Buffer.from(request.fcn ? request.fcn : 'invoke', 'utf8'));
  logger.debug('buildTransactionProposal - adding function arg:%s', request.fcn ? request.fcn : 'invoke');

  for (let i = 0; i < request.args.length; i++) {
    logger.debug('buildTransactionProposal - adding arg:%s', request.args[i]);
    args.push(Buffer.from(request.args[i], 'utf8'));
  }

  let invokeSpec = {
    type: 1, //_ccProto.ChaincodeSpec.Type.GOLANG,
    chaincode_id: {
      name: request.chaincodeId
    },
    input: {
      args: args
    }
  }

  var proposal, header;
  var channelHeader = clientUtils.buildChannelHeader(
    _commonProto.HeaderType.ENDORSER_TRANSACTION,
    request.chainId,
    request.txId.getTransactionID(),
    null,
    request.chaincodeId,
    clientUtils.buildCurrentTimestamp(),
    null
    // request.clientCertHash
  );
  header = clientUtils.buildHeader(identity, channelHeader, request.txId.getNonce());
  proposal = clientUtils.buildProposal(invokeSpec, header, request.transientMap);
  return proposal
}


/**
 * sign buffer with private key
 * @param  {String} private_key    [description]
 * @param  {Buffer} payload_bytes [description]
 * @return {Buffer}               [description]
 */
ClientProxy.prototype.sign = function (private_key, payload_bytes) {
  var key = new ECDSAKey(ClientProxy.getKey(private_key));
  // sign the proposal

  var digest = this._cryptoSuite.hash(payload_bytes)
  var sig = this._cryptoSuite.sign(key, digest)
  const signature = Buffer.from(sig);
  return signature
}

/**
 * verifyProposal
 * @param  {Buffer} signature
 * @param  {Buffer} proposal_bytes
 * @return {bool}
 */
ClientProxy.prototype.verifyProposal = function (signature, proposal_bytes) {
  var identity = this._identity
  return identity.verify(proposal_bytes, signature)
}

/**
 * send proposal to peer(s)
 * @param  {array|string} targets  peer target(s)
 * @param  {Buffer} proposal_bytes buffer of proposal
 * @param  {string} signature  buffer of proposal signature
 * @param  {Number} timeout
 * @param  {Object} opts  targets options  e.g. {pem:'--- tls cert ---'}
 * @return {[type]}                [description]
 */
ClientProxy.prototype.sendPeersProposal = function (targets, proposal_bytes, signature, timeout, opts) {
  var self = this
  var peer_targets
  if (Array.isArray(targets)) {
    peer_targets = targets.map(function (item) {
      return self._client.newPeer(item, opts)
    })

  } else {
    peer_targets = [self._client.newPeer(targets, opts)]
  }

  const signedProposal = {
    signature: signature,
    proposal_bytes: proposal_bytes
  };

  return clientUtils.sendPeersProposal(peer_targets, signedProposal, timeout)
    .then(function (responses) {

      var response_error = responses.filter(function (item) {
        return item instanceof Error
      })[0]

      if (response_error) {
        return Promise.reject(response_error);
      } else {
        return Promise.resolve(responses)
      }

    })

}

/**
 * [hasWriteSet description]
 * @param  {[type]}  proposalResponses [description]
 * @return {Boolean}                   [description]
 */
ClientProxy.prototype.hasWriteSet = function (proposalResponses) {
  return proposalResponses[0].rwset && proposalResponses[0].rwset.writes.length > 0
}

/**
 * [decodeProposal description]
 * @param  {[type]} proposal_bytes [description]
 * @return {[type]}                [description]
 */
ClientProxy.prototype.decodeProposal = function (proposal_bytes) {
  return blockDecoder.decodeProposal(proposal_bytes)
}



/**
 * [decodeProposalResponse description]
 * @param  {[type]} proposalResponses [description]
 * @return {[type]}                   [description]
 */
ClientProxy.prototype.decodeProposalResponse = function (proposalResponses) {
  var self = this
  for (var i in proposalResponses) {
    var proposalResponse = proposalResponses[i]
    if (proposalResponse.payload) {
      var payloadBytes = proposalResponse.payload.toBytes()
      proposalResponse.payload_decoded = blockDecoder.decodeProposalResponsePayload(payloadBytes)
      var extension = proposalResponse.payload_decoded.extension
      if (extension) {
        var chaincodeName = extension.chaincode_id.name
        proposalResponse.rwset = (extension.results.ns_rwset.filter(function (item) {
          return item.namespace == chaincodeName
        })[0] || {}).rwset
      }
    }

    logger.debug("proposalResponses[" + i + "]=", JSON.stringify(proposalResponse))
    if ( proposalResponse.response && proposalResponse.response.status === 200) {
      //logger.info('invoke chaincode proposal was good');
      proposalResponse.chaincode_result = formatChaincodeResult(proposalResponse.response.payload.toString())
    } else {

      //logger.error('invoke chaincode proposal was bad');
      // return Promise.reject(new Error(proposalResponses[i].details) )
      proposalResponse.chaincode_error = new Error(proposalResponse.response.message)
    }
  }
  return proposalResponses
}

/**
 * [buildTransactionPayload description]
 * @param  {[type]} proposal  [description]
 * @param  {[type]} proposalResponses [description]
 * @return {Buffer}                   [description]
 */
ClientProxy.prototype.buildTransactionPayload = function (proposal, proposalResponses) {
  var self = this;
  var errorMsg;

  if (!proposalResponses) {
    errorMsg = 'Missing "proposalResponses" parameter in transaction request';
  }
  if (!proposal) {
    errorMsg = 'Missing "proposal" parameter in transaction request';
  }
  var chaincodeProposal;

  if (proposal instanceof Buffer) {
    chaincodeProposal = _proposalProto.Proposal.decode(proposal);
  } else if (proposal instanceof Object) {
    chaincodeProposal = proposal;
  } else {
    errorMsg = '"proposal" parameter must be byte or object';
  }

  if (errorMsg) {
    logger.error('sendTransaction error ' + errorMsg);
    throw new Error(errorMsg);
  }

  const endorsements = [];
  let proposalResponse = proposalResponses;
  if (Array.isArray(proposalResponses)) {
    for (let i = 0; i < proposalResponses.length; i++) {
      // make sure only take the valid responses to set on the consolidated response object
      // to use in the transaction object
      if (proposalResponses[i].response && proposalResponses[i].response.status === 200) {
        proposalResponse = proposalResponses[i];
        endorsements.push(proposalResponse.endorsement);
      }
    }
  } else {
    if (proposalResponse && proposalResponse.response && proposalResponse.response.status === 200) {
      endorsements.push(proposalResponse.endorsement);
    }
  }

  if (endorsements.length < 1) {
    logger.error('sendTransaction - no valid endorsements found');
    throw new Error('no valid endorsements found');
  }

  const header = _commonProto.Header.decode(chaincodeProposal.getHeader());

  const chaincodeEndorsedAction = new _transProto.ChaincodeEndorsedAction();
  chaincodeEndorsedAction.setProposalResponsePayload(proposalResponse.payload);
  chaincodeEndorsedAction.setEndorsements(endorsements);

  const chaincodeActionPayload = new _transProto.ChaincodeActionPayload();
  chaincodeActionPayload.setAction(chaincodeEndorsedAction);

  // the TransientMap field inside the original proposal payload is only meant for the
  // endorsers to use from inside the chaincode. This must be taken out before sending
  // to the orderer, otherwise the transaction will be rejected by the validators when
  // it compares the proposal hash calculated by the endorsers and returned in the
  // proposal response, which was calculated without the TransientMap
  const originalChaincodeProposalPayload = _proposalProto.ChaincodeProposalPayload.decode(chaincodeProposal.payload);
  const chaincodeProposalPayloadNoTrans = new _proposalProto.ChaincodeProposalPayload();
  chaincodeProposalPayloadNoTrans.setInput(originalChaincodeProposalPayload.input); // only set the input field, skipping the TransientMap
  chaincodeActionPayload.setChaincodeProposalPayload(chaincodeProposalPayloadNoTrans.toBuffer());

  const transactionAction = new _transProto.TransactionAction();
  transactionAction.setHeader(header.getSignatureHeader());
  transactionAction.setPayload(chaincodeActionPayload.toBuffer());

  const actions = [];
  actions.push(transactionAction);

  const transaction = new _transProto.Transaction();
  transaction.setActions(actions);


  const payload = new _commonProto.Payload();
  payload.setHeader(header);
  payload.setData(transaction.toBuffer());

  const payload_bytes = payload.toBuffer();

  return payload_bytes
}

ClientProxy.prototype.getChannel = function (channel_id) {
  // return  this._client.getChannel(channel_id);
  var channel
  try {
    channel = this._client.newChannel(channel_id);
  } catch (e) {
    channel = this._client.getChannel(channel_id);
  }
  return channel
}

/**
 * send transaction to  orderer and wait for tx event
 * @param  {object} target  target Object  e.g. {orderer_url,event_url,orderer_opts,peer_opts}
 * @param  {Buffer} tx_payload_bytes [description]
 * @param  {Buffer} signature        [description]
 * @param  {int} timeout          [description]
 * @return {promise}  Promise of object  e.g. {tx_id:''}
 */
ClientProxy.prototype.sendTransaction = function (target, tx_payload_bytes, signature, timeout) {
  var self = this
  var orderer = self._client.newOrderer(target.orderer_url, target.orderer_opts)

  const envelope = {
    signature: signature,
    payload: tx_payload_bytes
  };

  var envelope_payload_decoded = blockDecoder.decodeEnvelopePayload(tx_payload_bytes)
  var channel_id = envelope_payload_decoded.header.channel_header.channel_id

  var channel = self.getChannel(channel_id);


  channel.addOrderer(orderer);
  // console.log('channel=',channel)
  var tx_id = envelope_payload_decoded.header.channel_header.tx_id

  return orderer.sendBroadcast(envelope)
    .then(function (result) {
      if (result.status == 'SUCCESS') {
        if (target.event_url) {
          var event_hub = self._client.newEventHub()
          event_hub.setPeerAddr(target.event_url, target.peer_opts)
          return self._txEventMonitor(tx_id, event_hub, timeout)
        } else {
          return Promise.resolve({tx_id: tx_id})
        }

      } else {
        return Promise.reject(new Error("sendTransactionError:" + result.status))
      }
    })
}

/**
 * [_txEventMonitor description]
 * @param  {[type]} tx_id     [description]
 * @param  {[type]} event_hub [description]
 * @param  {[type]} timeout   [description]
 * @return {[type]}           [description]
 */
ClientProxy.prototype._txEventMonitor = function (tx_id, event_hub, timeout) {
  return new Promise(function (resolve, reject) {
    //set timeout handle
    var event_timeout = setTimeout(function () {
      // do the housekeeping when there is a problem
      event_hub.unregisterTxEvent(tx_id);
      var error_message = 'Timeout - Failed to receive the transaction event.tx=' + tx_id
      logger.warn(error_message)
      reject(new Error(error_message));
    }, timeout)

    //registerTxEvent
    event_hub.registerTxEvent(
      tx_id,
      function (tx, code, block_num) {
        // logger.info('The chaincode invoke chaincode transaction has been committed on peer %s',eventHub.getPeerAddr());
        // logger.info('Transaction %s has status of %s in block %s', tx, code, block_num);
        clearTimeout(event_timeout);
        event_hub.unregisterTxEvent(tx_id);
        event_hub.disconnect();

        if (code !== 'VALID') {
          var message = util.format('The invoke chaincode transaction was invalid, code:%s', code);
          // logger.error(message);
          reject(new Error(message));
        } else {
          var message = 'The invoke chaincode transaction was valid.';
          // logger.info(message);
          resolve({
            tx_id: tx,
            block: block_num
          });
        }
      },
      function (error) {
        clearTimeout(event_timeout);
        logger.warn('Failed to receive the transaction event ::' + error);
        reject(error);
      },
      {unregister: true, disconnect: true}
    )
    //connect
    event_hub.connect()
  })
}



ClientProxy.prototype.decodeChaincodeInvocationSpec = function(decodeChaincodeInvocationSpec){
  return _chaincodeProto.ChaincodeInvocationSpec.decode(decodeChaincodeInvocationSpec)
}

module.exports = ClientProxy
