const jsrsasign = require('jsrsasign');

const crypto = require('crypto');
const ecies = require('standard-ecies');

var cyptoUtil = {}

//获取公钥或私钥
cyptoUtil.getKey = function(keyObj){
    var key = {}
    if(keyObj.hextype){
        key = jsrsasign.KEYUTIL.getKey(keyObj.key,keyObj.passcode,keyObj.hextype); 
    }
    else if(keyObj.passcode){
        key = jsrsasign.KEYUTIL.getKey(keyObj.key,keyObj.passcode); 
    }
    else{
        key = jsrsasign.KEYUTIL.getKey(keyObj.key); 
    }
    return key
}

/**
get PEM formatted private or public key file from a RSA/ECDSA/DSA key object 
{Object} keyObjOrHex key object RSAKey, KJUR.crypto.ECDSA or KJUR.crypto.DSA to encode to
{String} formatType (OPTION) output format type of "PKCS1PRV", "PKCS5PRV" or "PKCS8PRV" for private key
{String} passwd (OPTION) password to protect private key
{String} encAlg (OPTION) encryption algorithm for PKCS#5. currently supports DES-CBC, DES-EDE3-CBC and AES-{128,192,256}-CBC
{String} hexType (OPTION) type of hex string (ex. pkcs5prv, pkcs8prv)
{String} ivsaltHex hexadecimal string of IV and salt (default generated random IV)
*/
cyptoUtil.getPem = function(keyObj){
    if(keyObj.ivsaltHex){
        return  jsrsasign.KEYUTIL.getPEM(keyObj.keyObjOrHex, keyObj.formatType, keyObj.passwd, keyObj.encAlg, keyObj.hexType, keyObj.ivsaltHex)
    }
    else if(keyObj.hexType){
        return  jsrsasign.KEYUTIL.getPEM(keyObj.keyObjOrHex, keyObj.formatType, keyObj.passwd, keyObj.encAlg, keyObj.hexType)
    }
    else if(keyObj.encAlg){
        return  jsrsasign.KEYUTIL.getPEM(keyObj.keyObjOrHex, keyObj.formatType, keyObj.passwd, keyObj.encAlg)
    }
    else if(keyObj.passwd){
        return  jsrsasign.KEYUTIL.getPEM(keyObj.keyObjOrHex, keyObj.formatType, keyObj.passwd)
    }
    else if(keyObj.formatType){
        return  jsrsasign.KEYUTIL.getPEM(keyObj.keyObjOrHex, keyObj.formatType)
    }
    else{
        return  jsrsasign.KEYUTIL.getPEM(keyObj.keyObjOrHex)
    }
}

var RSA = {}

// keyLength e.g. 256 512
RSA.generateKeypair = function(keyLength){
    return jsrsasign.KEYUTIL.generateKeypair("RSA", keyLength);
}
/** RSA 公钥加密
 * @param keyObj  {JSON} 私钥对象。根据公钥形式的不同，内容如下：
 *  RSA private/public key 对象: keyObj={key:RSAKey对象}
 *  RSA public key 参数对象:  keyObj={key:{n: n, e: e}}
 *  X.509v1/v3 PEM 证书： keyObj={key:证书内容的字符串} 
 *  PKCS#8 十六进制 RSA 公钥: keyObj={key:公钥内容字符串,passcode:null,hextype:"pkcs8pub"} 
 *  PKCS#8 PEM形式的 RSA 公钥字符串: keyObj={key:公钥的pem内容字符串}
 * 
 * @param plaintext {String} 待就加密的明文 
 * @param opts {JSON} 选项对象 {algName:string} 其中algName 的值（默认为 RSA）可以为：
 *  RSA - RSA/ECB/PKCS1Padding (default for RSAKey)
 *  RSAOAEP - RSA/ECB/OAEPWithSHA-1AndMGF1Padding
 *  RSAOAEP224 - RSA/ECB/OAEPWithSHA-224AndMGF1Padding(*)
 *  RSAOAEP256 - RSA/ECB/OAEPWithSHA-256AndMGF1Padding
 *  RSAOAEP384 - RSA/ECB/OAEPWithSHA-384AndMGF1Padding(*)
 *  RSAOAEP512 - RSA/ECB/OAEPWithSHA-512AndMGF1Padding(*)
 * 
 * @return {String} 十六进制字符串的密文
*/
RSA.encryptByPub = function(keyObj, plaintext, opts){
    var pubObj = cyptoUtil.getKey(keyObj)
    opts = opts || {}
    opts.algName = opts.algName || "RSA"
	return jsrsasign.crypto.Cipher.encrypt(plaintext, pubObj, opts.algName)
}

/** RSA 私钥解密
 * @param keyObj  {JSON} 私钥对象。根据私钥形式的不同，内容如下：
 *  RSA private/public key 对象: keyObj={key:RSAKey对象}
 *  RSA private key 参数对象:  keyObj={key:{n: n, e: e, d: d, p: p, q: q, dp: dp, dq: dq, co: co}}
 *  X.509v1/v3 PEM 证书： keyObj={key:证书内容的字符串} 
 *  PKCS#5 十六进制 RSA private key: keyObj={key:私钥十六进制字符串,passcode:null,hextype:"pkcs5prv"} 
 *  PKCS#5 PEM RSA private key: keyObj={key:pem字符串}
 *  PKCS#8 PEM RSA private key: keyObj={key:pem字符串}
 *  PKCS#5 加密的 PEM RSA private key: keyObj={key:pem字符串,passcode:加密私钥的passcode} 
 *  PKCS#8 加密的 PEM RSA private key: keyObj={key:pem字符串,passcode:加密私钥的passcode} 
 * 
 * @param encryptedtext {Hex String} 待解密的十六进制 密文
 * @param opts {JSON} 选项对象 {algName:string} 其中algName（默认为RSA）的值可以为：
 *  RSA - RSA/ECB/PKCS1Padding (default for RSAKey)
 *  RSAOAEP - RSA/ECB/OAEPWithSHA-1AndMGF1Padding
 *  RSAOAEP224 - RSA/ECB/OAEPWithSHA-224AndMGF1Padding(*)
 *  RSAOAEP256 - RSA/ECB/OAEPWithSHA-256AndMGF1Padding
 *  RSAOAEP384 - RSA/ECB/OAEPWithSHA-384AndMGF1Padding(*)
 *  RSAOAEP512 - RSA/ECB/OAEPWithSHA-512AndMGF1Padding(*)
 * 
 * @return {String} 明文的buffer
*/
RSA.decryptByPrv = function(keyObj, encryptedtext, opts){
    var prvObj = cyptoUtil.getKey(keyObj)
    opts = opts || {}
    opts.algName = opts.algName || "RSA"
    return Buffer.from(jsrsasign.crypto.Cipher.decrypt(encryptedtext, prvObj, "RSA"),'utf8')
}

var ECC = {}

ECC.options = {
    hashName: 'sha256',
    hashLength: 32,
    macName: 'sha256',
    macLength: 32,
    curveName: 'prime256v1',
    symmetricCypherName: 'aes-256-ecb',
    iv: null, // iv is used in symmetric cipher, set null if cipher is in ECB mode. 
    keyFormat: 'uncompressed',
    s1: null, // optional shared information1
    s2: null // optional shared information2
}

// curveName e.g. secp256r1, secp256k1 and secp384r1
ECC.generateKeypair = function(curveName){
    return jsrsasign.KEYUTIL.generateKeypair("EC", curveName);
}

/** ECC 证书中提取公钥加密 文本 
 * @param keyObj  {JSON} 私钥对象。根据公钥形式的不同，内容如下：
 *  ECC private/public key 对象: keyObj={key:KJUR.crypto.ECDSA 对象}
 *  ECC public key 参数对象:  keyObj={key:{xy: xy, curve: curveName}}
 *  X.509v1/v3 PEM 证书： keyObj={key:证书内容的字符串} 
 *  PKCS#8 十六进制 ECC 公钥: keyObj={key:公钥内容字符串,passcode:null,hextype:"pkcs8pub"} 
 *  PKCS#8 PEM形式的 ECC 公钥字符串: keyObj={key:公钥的pem内容字符串}
 * 
 * @param plaintext {String} 待就加密的明文 
 * @param opts {JSON} 选项对象 默认值为
 * {
    hashName: 'sha256',
    hashLength: 32,
    macName: 'sha256',
    macLength: 32,
    curveName: 'prime256v1',
    symmetricCypherName: 'aes-256-ecb',
    iv: null, // iv is used in symmetric cipher, set null if cipher is in ECB mode. 
    keyFormat: 'uncompressed',
    s1: null, // optional shared information1
    s2: null // optional shared information2
    } 
 * 
 * @return {String} 十六进制字符串的密文
*/
ECC.encryptByPub = function(keyObj,plaintext,opts=ECC.options){
    var pubObj = cyptoUtil.getKey(keyObj)
    var plainText = new Buffer(plaintext);
    var encryptedText = ecies.encrypt(Buffer.from(pubObj.pubKeyHex,'hex'), plainText, opts); 			
	return encryptedText.toString("hex")
}


/** ECC 私钥解密
 * @param keyObj  {JSON} 私钥对象。根据私钥形式的不同，内容如下：
 *  ECC private/public key 对象: keyObj={key:KJUR.crypto.ECDSA 对象}
 *  ECC private key 参数对象:  keyObj={key:{d: d, curve: curveName}}
 *  X.509v1/v3 PEM 证书： keyObj={key:证书内容的字符串} 
 *  PKCS#8 PEM ECDSA private key: keyObj={key:pem字符串}
 *  PKCS#8 加密的 PEM ECDSA private key: keyObj={key:pem字符串,passcode:加密私钥的passcode} 
 * 
 * @param encryptedtext {Hex String} 待解密的密文
 *
 * @param opts {JSON} 选项对象 默认值为
 * {
    hashName: 'sha256',
    hashLength: 32,
    macName: 'sha256',
    macLength: 32,
    curveName: 'prime256v1',
    symmetricCypherName: 'aes-256-ecb',
    iv: null, // iv is used in symmetric cipher, set null if cipher is in ECB mode. 
    keyFormat: 'uncompressed',
    s1: null, // optional shared information1
    s2: null // optional shared information2
    } 
 * 
 * @return {Buffer} 明文的buffer
*/
ECC.decryptByPrv = function(keyObj,encryptedText,opts=ECC.options){
    var prvObj = cyptoUtil.getKey(keyObj)
    var ecdh = crypto.createECDH(opts.curveName);
    ecdh.setPrivateKey(Buffer.from(prvObj.prvKeyHex,'hex'));
    var decryptedText = ecies.decrypt(ecdh,Buffer.from(encryptedText,"hex"), opts);			
	return decryptedText
}

var DES = {}
DES.options = {
    algName: 'des-ede3',    //3des-ecb
    autoPad: true,
    iv: null
}

/** des 加密
 * @param keyObj  {JSON} 密钥对象。{key：utf8 string}
 * 
 * @param plaintext {String} 待就加密的明文 
 * @param opts {JSON} 选项对象 默认值为3des
 * {
    algName: 'des-ede3',    //3des-ecb
    autoPad: true,
    iv: null
    }
 * 
 * @return {Hex String} 十六进制字符串的密文
*/
DES.encrypt = function(keyObj,plaintext,opts=DES.options) {
    var key = new Buffer(keyObj.key);
    var iv = new Buffer(opts.iv ? opts.iv : 0)
    var algName = opts.algName
    var autoPad = opts.autoPad

    //encrypt
    var cipher = crypto.createCipheriv(algName, key, iv);
    cipher.setAutoPadding(autoPad)  //default true
    var encryptHexText = cipher.update(plaintext, 'utf8', 'hex');
    encryptHexText += cipher.final('hex');
    // console.log(algName, encryptHexText)
    return encryptHexText;
}

//传入 hex输出utf8

/** des 解密
 * @param keyObj  {JSON} 密钥对象。{key：utf8 string}
 * 
 * @param encryptedText {Hex String} 待解密的十六进制密文
 * @param opts {JSON} 选项 默认值为3des
 * {
    algName: 'des-ede3',    //3des-ecb
    autoPad: true,
    iv: null
    }
 * 
 * @return {Buffer} 明文的buffer
*/
DES.decrypt = function(keyObj,encryptedText,opts=DES.options) {  
    var key = new Buffer(keyObj.key);  
    var iv = new Buffer(opts.iv ? opts.iv : 0)
    var algName = opts.algName  
    var autoPad = opts.autoPad 
    //decrypt
    var decipher = crypto.createDecipheriv(algName, key, iv);  
    decipher.setAutoPadding(autoPad)  
    var txt = decipher.update(encryptedText, 'hex', 'utf8');  
    txt += decipher.final('utf8');      
    return Buffer.from(txt,"utf8");
}

/** 3DES DES RSA ECC 加密
 * @param keyObj {JSON} 请参考 DES.encrypt,RSA.encrypt,ECC.encrypt 函数
 * @param plaintext {String} 待就加密的明文
 * @param  opts {JSON} 选项 格式为
    {
        type:String, // 加密类型 可选值为 DES|RSA|ECC
        opts:加密类型所需的选项，此项可选，默认值对应为 3DES | RSA/ECB/PKCS1Padding | ECC secp256r1曲线/sha256/aes-256-ecb
    }
 * @return {Hex String} 十六进制字符串的密文
*/
cyptoUtil.encrypt = function(keyObj, plaintext, opts){
    if(opts.type == 'DES'){
        opts.opts = opts.opts || DES.opts
        return DES.encrypt(keyObj, plaintext, opts.opts)
    }
    else if(opts.type == 'ECC'){
        opts.opts = opts.opts || ECC.opts
        return ECC.encryptByPub(keyObj, plaintext, opts.opts)
    }
    else if(opts.type == 'RSA'){
        opts.opts = opts.opts || RSA.opts
        return RSA.encryptByPub(keyObj, plaintext, opts.opts)
    }
    else{
        throw new Error("Unknown Cipher Type!")
    }
}

/** 3DES DES RSA ECC 解密
 * @param keyObj {JSON} 请参考 DES.decrypt,RSA.decrypt,ECC.decrypt 函数
 * @param plaintext {Hex String} 待解密十六进制密文
 * @param  opts {JSON} 选项 格式为
    {
        type:String, // 加密类型 可选值为 DES|RSA|ECC
        opts:加密类型所需的选项，此项可选，默认值对应为 3DES | RSA/ECB/PKCS1Padding | ECC secp256r1曲线/sha256/aes-256-ecb
    }
 * @return {Buffer} 明文的buffer
*/
cyptoUtil.decrypt = function(keyObj, encryptedText, opts){
    if(opts.type == 'DES'){
        opts.opts = opts.opts || DES.opts
        return DES.decrypt(keyObj, encryptedText,opts.opts)
    }
    else if(opts.type == 'ECC'){
        opts.opts = opts.opts || ECC.opts
        return ECC.decryptByPrv(keyObj, encryptedText,opts.opts)
    }
    else if(opts.type == 'RSA'){
        opts.opts = opts.opts || RSA.opts
        return RSA.decryptByPrv(keyObj, encryptedText,opts.opts)
    }
    else{
        throw new Error("Unknown Cipher Type!")
    }
}

module.exports = cyptoUtil
module.exports.RSA = RSA
module.exports.ECC = ECC
module.exports.DES = DES