/**
 * Copyright 2016 IBM All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

/**
 * Implement hash primitives.
 */
const logger = require('./utils').getLogger('hash');
const jsSHA3 = require('js-sha3');
const { sha3_256, sha3_384, shake_256 } = jsSHA3;
const crypto = require('crypto');

class hashBaseClass {
	constructor() {
		this.reset();
	}

	hash(data) {
		return this.reset().update(data).finalize();
	}
	reset() {
		return this;
	}
	update(data) {
		logger.debug(`update(${typeof data})`);
		this._hash.update(data);
		return this;
	}
	finalize() {
	}
}
class hash_sha2_256 extends hashBaseClass {
	constructor() {
		super();
		this.blockSize = 512;
	}
	reset() {
		this._hash = crypto.createHash('sha256');
		return super.reset();
	}
	finalize() {
		const hash = this._hash.digest('hex');
		this.reset();
		return hash;
	}
}
class hash_sha2_384 extends hashBaseClass {
	constructor() {
		super();
		this.blockSize = 1024;
	}
	reset() {
		this._hash = crypto.createHash('sha384');
		return super.reset();
	}
	finalize() {
		const hash = this._hash.digest('hex');
		this.reset();
		return hash;
	}
}
class hash_sha3_256 extends hashBaseClass {
	static hashSimple(data) {
		return sha3_256(data);
	}
	constructor() {
		super();
		this.blockSize = 1088;
	}
	reset() {
		this._hash = sha3_256.create();
		return super.reset();
	}
	finalize() {
		const hash = this._hash.hex();
		this.reset();
		return hash;
	}

}
class hash_sha3_384 extends hashBaseClass {
	static hashSimple(data) {
		return sha3_384(data);
	}
	constructor() {
		super();
		this.blockSize = 832;
	}
	reset() {
		this._hash = sha3_384.create();
		return super.reset();
	}
	finalize() {
		const hash = this._hash.hex();
		this.reset();
		return hash;
	}
}

exports.hash_sha3_256 = hash_sha3_256;
exports.hash_sha3_384 = hash_sha3_384;
exports.hash_sha2_256 = hash_sha2_256;
exports.hash_sha2_384 = hash_sha2_384;
exports.sha2_256 = (data) => {
	return (new hash_sha2_256()).hash(data);
};
exports.sha3_256 = sha3_256;
exports.sha2_384 = (data) => {
	return (new hash_sha2_384()).hash(data);
};
exports.sha3_384 = sha3_384;
exports.shake_256 = shake_256;//TODO
